"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.$unitDefinition = exports.$removedefault = exports.$setdefault = exports.$default = exports.$config = exports.$currency = exports.$continentName = exports.$language = exports.$country = exports.$locales = exports.$tmp = exports.TSDefaults = void 0;
const commons_1 = require("./commons");
const types_1 = require("./types");
const fs_1 = require("./fs");
const os_1 = __importDefault(require("os"));
const tscountry_1 = require("./tscountry");
const locales_json_1 = __importDefault(require("./locales.json"));
const utils_1 = require("./utils");
const env_1 = require("./env");
const strings_1 = require("./strings");
const tserrors_1 = require("./tserrors");
class TSDefaults {
    constructor() {
        this._tmpDirectory = (0, utils_1.$inbrowser)() ? '' : os_1.default.tmpdir();
        this._values = {};
        this._localizations = {};
        this._countriesMap = new Map();
        Object.keys(types_1.Countries).forEach(e => this._countriesMap.set(e, e));
        this._languagesMap = new Map();
        Object.keys(types_1.Languages).forEach(e => this._languagesMap.set(e, e));
        this._currenciesMap = new Map();
        Object.keys(types_1.Currencies).forEach(e => this._currenciesMap.set(e, e));
        this._managedLanguages = TSDefaults.__locales.map(loc => loc.language);
        let managedLocales = new Map();
        TSDefaults.__locales.forEach(loc => {
            managedLocales.set(loc.language, loc);
            this._managedLanguages.forEach(l => managedLocales.set((0, strings_1.$ascii)(loc.names[l].toLowerCase()), loc));
        });
        this._managedLocalesMap = managedLocales;
        const cmap = tscountry_1.TSCountry.loadCountries(managedLocales, this._managedLanguages);
        const c = cmap.get(types_1.Countries.FR);
        if (!(0, commons_1.$ok)(c)) {
            tserrors_1.TSError.throw(`Did not find country '${types_1.Countries.FR}' to be the default`);
        }
        this._defaultCountry = c;
        this._defaultCurrency = c.currency;
        this._defaultLanguage = c.locales.language;
    }
    managedLanguages() { return [...this._managedLanguages]; }
    managedLanguage(s) {
        if (!(0, commons_1.$ok)(s)) {
            return this.defaultLanguage;
        }
        const v = (0, strings_1.$ascii)((0, strings_1.$ftrim)(s).toLowerCase());
        const locales = this._managedLocalesMap.get(v);
        if ((0, commons_1.$ok)(locales)) {
            return locales.language;
        }
        return null;
    }
    continentName(cont, lang) {
        if (!(0, commons_1.$ok)(cont)) {
            return null;
        }
        const locales = this._managedLocalesMap.get((0, commons_1.$value)(lang, this.defaultLanguage));
        return (0, commons_1.$valueornull)(locales === null || locales === void 0 ? void 0 : locales.continentNames[cont]);
    }
    country(s) {
        if ((0, commons_1.$ok)(s) && !(0, commons_1.$isstring)(s)) {
            return null;
        }
        const v = (0, strings_1.$ascii)((0, strings_1.$ftrim)(s).toUpperCase());
        const managedCountry = tscountry_1.TSCountry.country(v);
        if ((0, commons_1.$ok)(managedCountry)) {
            return managedCountry.alpha2Code;
        }
        return (0, commons_1.$valueornull)(this._countriesMap.get(v));
    }
    language(s) {
        if (!(0, commons_1.$ok)(s)) {
            return this.defaultLanguage;
        }
        if (s instanceof tscountry_1.TSCountry) {
            return s.language;
        }
        const v = (0, strings_1.$ascii)((0, strings_1.$ftrim)(s).toLowerCase());
        const locales = this._managedLocalesMap.get(v);
        if ((0, commons_1.$ok)(locales)) {
            return locales.language;
        }
        return (0, commons_1.$valueornull)(this._languagesMap.get(v));
    }
    currency(s) {
        if (!(0, commons_1.$ok)(s)) {
            return this.defaultCurrency;
        }
        if (s instanceof tscountry_1.TSCountry) {
            return s.currency;
        }
        s = (0, strings_1.$ascii)((0, strings_1.$ftrim)(s).toUpperCase());
        const c = tscountry_1.TSCountry.country(s);
        if ((0, commons_1.$ok)(c)) {
            return c.currency;
        }
        return (0, commons_1.$valueornull)(this._currenciesMap.get(s));
    }
    addLocalizations(lang, loc) {
        if (!this._managedLocalesMap.get(lang)) {
            tserrors_1.TSError.throw(`Impossible to add localizations to non managed language '${lang}'`, { language: lang, localizations: loc });
        }
        else if (!(0, commons_1.$isobject)(loc)) {
            tserrors_1.TSError.throw(`Impossible to add non string-dictionary localization to language '${lang}'`, { language: lang, localizations: loc });
        }
        let actualLocalization = this._localizations[lang];
        if (!(0, commons_1.$ok)(actualLocalization)) {
            this._localizations[lang] = Object.assign({}, loc);
        }
        else {
            this._localizations[lang] = Object.assign(Object.assign({}, actualLocalization), loc);
        }
    }
    localizations(locale) {
        if (locale instanceof tscountry_1.TSCountry) {
            return this._localizations[locale.language] || {};
        }
        locale = (0, strings_1.$ftrim)(locale);
        if (locale.length) {
            const lang = this.language(locale);
            if ((0, commons_1.$ok)(lang)) {
                return this._localizations[lang] || {};
            }
            const c = tscountry_1.TSCountry.country(locale);
            if ((0, commons_1.$ok)(c)) {
                return this._localizations[c.language] || {};
            }
        }
        return {};
    }
    locales(locale) {
        if (locale instanceof tscountry_1.TSCountry) {
            return locale.locales;
        }
        locale = (0, strings_1.$ascii)((0, strings_1.$ftrim)(locale));
        if (locale.length) {
            const locales = this._managedLocalesMap.get(locale.toLowerCase());
            if ((0, commons_1.$ok)(locales)) {
                return locales;
            }
            const c = tscountry_1.TSCountry.country(locale);
            if ((0, commons_1.$ok)(c)) {
                return c.locales;
            }
        }
        return this._managedLocalesMap.get(this.defaultLanguage);
    }
    get defaultCountry() { return this._defaultCountry; }
    setDefaultCountry(c) {
        if ((0, commons_1.$ok)(c)) {
            this._defaultCountry = c;
        }
        return this._defaultCountry;
    }
    get defaultLanguage() { return this._defaultLanguage; }
    setDefaultLanguage(l) {
        if ((0, commons_1.$ok)(this._managedLocalesMap.get(l))) {
            this._defaultLanguage = l;
        }
        return this._defaultLanguage;
    }
    get defaultCurrency() { return this._defaultCurrency; }
    setDefaultCurrency(c) {
        if ((0, commons_1.$ok)(this._currenciesMap.get(c))) {
            this._defaultCurrency = c;
        }
        return this._defaultCurrency;
    }
    get tmpDirectory() {
        tserrors_1.TSError.assertNotInBrowser('TSDefaults.setTmpDirectory');
        return this._tmpDirectory;
    }
    setTmpDirectory(path) {
        tserrors_1.TSError.assertNotInBrowser('TSDefaults.setTmpDirectory');
        if (!(0, fs_1.$isdirectory)(path)) {
            tserrors_1.TSError.throw('TSDefaults.setTmpDirectory(): bad path parameter', { path: path });
        }
        this._tmpDirectory = path;
    }
    setValue(key, value) {
        if ((0, commons_1.$length)(key)) {
            if ((0, commons_1.$ok)(value)) {
                this._values[key] = value;
            }
            else if ((0, commons_1.$ok)(this._values[key])) {
                delete this._values[key];
            }
        }
    }
    getValue(key) {
        const v = this._values[key];
        if ((0, commons_1.$ok)(v)) {
            return v;
        }
        return !(0, utils_1.$inbrowser)() && (0, commons_1.$isobject)(process) && (0, commons_1.$isobject)(process.env) ? process.env[key] : undefined;
    }
    configure(path, opts) {
        tserrors_1.TSError.assertNotInBrowser('TSDefaults.configure');
        if (!(0, commons_1.$isobject)(process) || !(0, commons_1.$isobject)(process.env)) {
            tserrors_1.TSError.throw('TSDefaults.configure(): process.env is not defined', { path: path, options: opts });
        }
        const debug = !!(opts === null || opts === void 0 ? void 0 : opts.debug);
        path = (0, fs_1.$absolute)((0, commons_1.$length)(path) > 0 ? path : '.env');
        const buffer = (0, fs_1.$readBuffer)(path);
        if (!(0, commons_1.$ok)(buffer)) {
            if (debug) {
                (0, utils_1.$logterm)(`&R&w TSDefaults.configure(): impossible to read environment file '${path}'  &0`);
            }
        }
        else if (!(0, commons_1.$ok)((0, env_1.$env)(buffer, Object.assign({ merge: process.env }, opts)))) {
            if (debug) {
                (0, utils_1.$logterm)(`&R&w TSDefaults.configure(): impossible to interpret environment file '${path}'  &0`);
            }
        }
    }
    static defaults() {
        if (!this.__instance) {
            this.__instance = new TSDefaults();
        }
        return this.__instance;
    }
}
exports.TSDefaults = TSDefaults;
TSDefaults.__locales = locales_json_1.default;
function $tmp() { return TSDefaults.defaults().tmpDirectory; }
exports.$tmp = $tmp;
function $locales(locale) { return TSDefaults.defaults().locales(locale); }
exports.$locales = $locales;
function $country(s) { return TSDefaults.defaults().country(s); }
exports.$country = $country;
function $language(s) { return TSDefaults.defaults().language(s); }
exports.$language = $language;
function $continentName(cont, lang) { return TSDefaults.defaults().continentName(cont, lang); }
exports.$continentName = $continentName;
function $currency(s) { return TSDefaults.defaults().currency(s); }
exports.$currency = $currency;
function $config(path, opts) { TSDefaults.defaults().configure(path, opts); }
exports.$config = $config;
function $default(key) { return TSDefaults.defaults().getValue(key); }
exports.$default = $default;
function $setdefault(key, value = undefined) { TSDefaults.defaults().setValue(key, value); }
exports.$setdefault = $setdefault;
function $removedefault(key) { TSDefaults.defaults().setValue(key, undefined); }
exports.$removedefault = $removedefault;
function $unitDefinition(key, locale) {
    const loc = !(0, commons_1.$ok)(locale) || (0, commons_1.$isstring)(locale) || (locale instanceof tscountry_1.TSCountry) ? $locales(locale) : locale;
    const a = loc.unitNames[key];
    if ((0, commons_1.$ok)(a)) {
        if ((0, commons_1.$isarray)(a)) {
            if ((0, commons_1.$isstring)(a[0]) && (0, commons_1.$isstring)(a[1])) {
                const s = (0, strings_1.$ftrim)(a[0]);
                const p = (0, strings_1.$ftrim)(a[1]);
                if (s.length > 0 && p.length > 0) {
                    return { singular: s, plural: p, unit: (0, commons_1.$isstring)(a[2]) ? (0, strings_1.$ftrim)(a[2]) : '' };
                }
            }
        }
        return null;
    }
    return undefined;
}
exports.$unitDefinition = $unitDefinition;
//# sourceMappingURL=tsdefaults.js.map