"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.$validateParsedValue = exports.$bool = exports.isLeafParserType = exports.TSParser = exports.TSCase = exports.TSParserActionContext = void 0;
const types_1 = require("./types");
const commons_1 = require("./commons");
const data_1 = require("./data");
const strings_1 = require("./strings");
const tscharset_1 = require("./tscharset");
const tscolor_1 = require("./tscolor");
const tscountry_1 = require("./tscountry");
const tsdate_1 = require("./tsdate");
const tsgeometry_1 = require("./tsgeometry");
const tsphonenumber_1 = require("./tsphonenumber");
const tsurl_1 = require("./tsurl");
const tserrors_1 = require("./tserrors");
const utils_1 = require("./utils");
var TSParserActionContext;
(function (TSParserActionContext) {
    TSParserActionContext["url"] = "url";
    TSParserActionContext["json"] = "json";
    TSParserActionContext["vargs"] = "vargs";
    TSParserActionContext["env"] = "env";
})(TSParserActionContext = exports.TSParserActionContext || (exports.TSParserActionContext = {}));
;
var TSCase;
(function (TSCase) {
    TSCase["standard"] = "standard";
    TSCase["lowercase"] = "lowercase";
    TSCase["uppercase"] = "uppercase";
})(TSCase = exports.TSCase || (exports.TSCase = {}));
class TSParser {
    static define(rootNode, errors) {
        return (0, commons_1.$valueornull)(_structureConstruction(rootNode, errors));
    }
    get isValid() { return this.errors.length === 0; }
    validate(value, opts) {
        return this._validate(value, '', opts);
    }
    interpret(value, opts) {
        return this._validate(value, '', opts) ? this.rawInterpret(value, opts) : null;
    }
    encode(value, opts) {
        return this._validate(value, '', opts) ? this.rawEncode(value, opts) : null;
    }
    parse(source, errors) {
        const opts = { errors: errors, context: TSParserActionContext.json };
        let value = undefined;
        const s = (0, strings_1.$trim)(source);
        if (s.length) {
            try {
                value = JSON.parse(s);
            }
            catch (_a) {
                value = undefined;
                _serror(opts, 'JSON', '.parse(stringSource) did fail');
            }
        }
        else {
            _serror(opts, 'stringSource', ' is empty');
        }
        return this.interpret(value, opts);
    }
    stringify(value, spaces = undefined, errors) {
        const opts = { errors: errors, context: TSParserActionContext.json };
        if (this._validate(value, '', opts)) {
            try {
                return JSON.stringify(this.rawEncode(value, opts), undefined, spaces);
            }
            catch (_a) {
                return null;
            }
        }
        return null;
    }
    toString() { return (0, utils_1.$inspect)(this.toJSON(), 15); }
    includedDefaultValue() { return undefined; }
    _validate(value, path, opts) {
        if (this.mandatory && !(0, commons_1.$ok)(value)) {
            return _serror(opts, path, 'is mandatory');
        }
        return true;
    }
    constructor(mandatory, checker, transformer, natifier) {
        this.errors = [];
        this._check = (_) => true;
        this._transform = (v) => v;
        this._natify = undefined;
        this.mandatory = mandatory;
        if ((0, commons_1.$isfunction)(checker)) {
            this._check = checker;
        }
        if ((0, commons_1.$isfunction)(transformer)) {
            this._transform = transformer;
        }
        if ((0, commons_1.$isfunction)(natifier)) {
            this._natify = natifier;
        }
    }
    ;
}
exports.TSParser = TSParser;
class TSDictionaryParser extends TSParser {
    constructor(node, keysParser, valuesParser) {
        const m = node._mandatory;
        super(!!m, node._checker, node._transformer, node._natifier);
        if (!(keysParser instanceof TSLeafParser) || !keysParser.isKey) {
            this.errors.push('_keysType field is not a valid node key definition');
        }
        if ((0, commons_1.$ok)(keysParser.includedDefaultValue()) || (0, commons_1.$ok)(valuesParser.includedDefaultValue())) {
            this.errors.push(`_keysParser or _valuesParser cannot contain a default value`);
        }
        if ((0, commons_1.$defined)(m) && !(0, commons_1.$isbool)(m)) {
            this.errors.push(`_mandatory field should be a boolean`);
        }
        if ((0, commons_1.$ok)(node._keysCase)) {
            this.errors.push(`_keysCase field should not be set for a dictionary`);
        }
        this._keysParser = keysParser;
        this._valuesParser = valuesParser;
    }
    nodeType() { return 'object'; }
    toJSON() {
        return {
            _type: 'dictionary',
            _mandatory: this.mandatory,
            _check: this._check,
            _transform: this._transform,
            _natify: this._natify,
            _keysStruct: this._keysParser.toJSON(),
            _valuesStruct: this._valuesParser.toJSON()
        };
    }
    _validate(value, path, opts) {
        if (!super._validate(value, path, opts)) {
            return false;
        }
        if ((0, commons_1.$ok)(value) && ((0, commons_1.$isarray)(value) || !(0, commons_1.$isobject)(value))) {
            return _serror(opts, path, 'is not a dictionary');
        }
        const entries = Object.entries(value);
        if (!(entries === null || entries === void 0 ? void 0 : entries.length) && this.mandatory) {
            return _serror(opts, path, ` is empty`);
        }
        let doomed = false;
        const p = path.length ? path : 'dict';
        for (let [k, v] of entries) {
            if (!this._keysParser._validate(k, `${p}.${k}[key]`, opts)) {
                doomed = true;
            }
            if (!this._valuesParser._validate(v, `${p}.${k}`, opts)) {
                doomed = true;
            }
        }
        if (doomed) {
            return false;
        }
        return this._check(value, opts) ? true : _serror(opts, path, 'specific check did fail');
    }
    rawInterpret(value, opts) {
        const ret = {};
        if (!(0, commons_1.$ok)(value) && !this.mandatory) {
            return value;
        }
        const entries = Object.entries(value);
        for (let [k, v] of entries) {
            const ki = this._keysParser.rawEncode(k, opts);
            const vi = this._valuesParser.rawInterpret(v, opts);
            if ((0, commons_1.$ok)(ki) && (0, commons_1.$ok)(vi)) {
                ret[ki] = vi;
            }
        }
        return this._transform(ret, opts);
    }
    rawEncode(value, opts) {
        const ret = {};
        const entries = Object.entries(value);
        for (let [k, v] of entries) {
            const ki = this._keysParser.rawEncode(k, opts);
            const vi = this._valuesParser.rawEncode(v, opts);
            if ((0, commons_1.$ok)(ki) && (0, commons_1.$ok)(vi)) {
                ret[ki] = vi;
            }
        }
        return this._natify ? this._natify(ret, opts) : ret;
    }
}
class TSObjectParser extends TSParser {
    constructor(node, itemsParsers, acceptsUncheckedItems) {
        const m = node._mandatory;
        super(!!m, node._checker, node._transformer, node._natifier);
        this._keyTransform = (s) => s;
        this._keysCase = TSCase.standard;
        this._defaultValueKeys = [];
        if ((0, commons_1.$defined)(m) && !(0, commons_1.$isbool)(m)) {
            this.errors.push(`_mandatory field should be a boolean`);
        }
        const l = (0, commons_1.$value)(node._keysCase, TSCase.standard);
        const kt = InternalCasingMap[l];
        if (!(0, commons_1.$defined)(kt)) {
            this.errors.push(`_keysCase field should one of ${Array.from((0, commons_1.$keys)(InternalCasingMap))}`);
        }
        else {
            this._keyTransform = kt;
            this._keysCase = l;
        }
        this._acceptsUncheckedItems = acceptsUncheckedItems;
        this._itemsParsers = itemsParsers;
        this._count = (0, commons_1.$objectcount)(itemsParsers);
        this._aliasUnsensitive = !!node._aliasUnsensitive;
        const aliases = node._aliases;
        if (aliases === null || aliases === void 0 ? void 0 : aliases.size) {
            if (this._aliasUnsensitive) {
                this._aliases = new Map();
                for (let [ak, tk] of aliases) {
                    ak = ak.toLowerCase();
                    if (this._aliases.get(ak)) {
                        this.errors.push(`_aliases field should not contain insensitive similar alias ${ak} twice.`);
                    }
                    this._aliases.set(ak, tk);
                }
            }
            else {
                this._aliases = new Map(aliases);
            }
        }
        else {
            this._aliases = new Map();
        }
        this._outputAliases = (0, commons_1.$ok)(node._outputAliases) ? new Map(node._outputAliases) : new Map();
        const entries = Object.entries(itemsParsers);
        for (let [k, p] of entries) {
            if ((0, commons_1.$ok)(p.includedDefaultValue())) {
                this._defaultValueKeys.push(k);
            }
        }
    }
    nodeType() { return 'object'; }
    toJSON() {
        const ret = {
            _type: 'object',
            _mandatory: this.mandatory,
            _check: this._check,
            _transform: this._transform,
            _natify: this._natify,
            _acceptsUncheckedItems: this._acceptsUncheckedItems
        };
        if (this._keysCase !== 'standard') {
            ret._keysCase = this._keysCase;
        }
        if (this._aliases.size > 0) {
            ret._aliases = this._aliases;
        }
        if (this._outputAliases.size > 0) {
            ret._outputAliases = this._outputAliases;
        }
        const structEntries = Object.entries(this._itemsParsers);
        for (let [k, struct] of structEntries) {
            ret[k] = struct.toJSON();
        }
        return ret;
    }
    _validate(value, path, opts) {
        if (!super._validate(value, path, opts)) {
            return false;
        }
        if ((0, commons_1.$ok)(value) && ((0, commons_1.$isarray)(value) || !(0, commons_1.$isobject)(value))) {
            return _serror(opts, path, 'is not an object');
        }
        const p = path.length ? path : 'object';
        const entries = Object.entries(value);
        const founds = new Set();
        let doomed = false;
        const atl = this._aliasUnsensitive;
        for (let [key, v] of entries) {
            const k = this._keyTransform((0, commons_1.$value)(this._aliases.get(atl ? key.toLowerCase() : key), key));
            if (founds.has(k)) {
                _serror(opts, path, `.${k} is present several times`);
                doomed = true;
                continue;
            }
            const struct = this._itemsParsers[k];
            if (!(0, commons_1.$ok)(struct)) {
                if (!this._acceptsUncheckedItems) {
                    _serror(opts, path, `.${k} is unknown`);
                    doomed = true;
                }
                continue;
            }
            if (!struct._validate(v, `${p}.${k}`, opts)) {
                doomed = true;
                continue;
            }
            founds.add(k);
        }
        if (founds.size < this._count) {
            const structEntries = Object.entries(this._itemsParsers);
            for (let [k, struct] of structEntries) {
                if (!founds.has(k) && struct.mandatory) {
                    _serror(opts, path, `.${k} is mandatory`);
                    doomed = true;
                }
            }
        }
        if (doomed) {
            return false;
        }
        return this._check(value, opts) ? true : _serror(opts, path, 'specific check did fail');
    }
    rawInterpret(value, opts) {
        if (!(0, commons_1.$ok)(value) && !this.mandatory) {
            return value;
        }
        const ret = {};
        const entries = Object.entries(value);
        const atl = this._aliasUnsensitive;
        for (let [key, v] of entries) {
            const k = this._keyTransform((0, commons_1.$value)(this._aliases.get(atl ? key.toLowerCase() : key), key));
            const struct = this._itemsParsers[k];
            if ((0, commons_1.$ok)(struct)) {
                const res = struct.rawInterpret(v, opts);
                if ((0, commons_1.$ok)(res)) {
                    ret[k] = res;
                }
            }
            else {
                ret[k] = v;
            }
        }
        for (let k of this._defaultValueKeys) {
            const kf = this._keyTransform((0, commons_1.$value)(this._aliases.get(atl ? k.toLowerCase() : k), k));
            if (!(0, commons_1.$ok)(ret[kf])) {
                ret[kf] = this._itemsParsers[k].includedDefaultValue();
            }
        }
        return this._transform(ret, opts);
    }
    rawEncode(value, opts) {
        const ret = {};
        const entries = Object.entries(value);
        const atl = this._aliasUnsensitive;
        for (let [key, v] of entries) {
            const k = this._keyTransform((0, commons_1.$value)(this._aliases.get(atl ? key.toLowerCase() : key), key));
            const struct = this._itemsParsers[k];
            const res = struct.rawEncode(v);
            if ((0, commons_1.$ok)(res)) {
                ret[(0, commons_1.$value)(this._outputAliases.get(k), k)] = res;
            }
        }
        return this._natify ? this._natify(ret, opts) : ret;
    }
}
class TSArrayParser extends TSParser {
    constructor(node, itemParser) {
        const m = node._mandatory;
        let min = 0, max = types_1.UINT_MAX;
        const errors = [];
        if ((0, commons_1.$ok)(node._min)) {
            if (!(0, commons_1.$isunsigned)(node._min)) {
                errors.push(`wrong minimal size definition for array node`);
            }
            else {
                min = (0, commons_1.$unsigned)(node._min);
            }
        }
        if ((0, commons_1.$ok)(node._max)) {
            if (!(0, commons_1.$isunsigned)(node._max)) {
                errors.push(`wrong maximal size definition for array node`);
            }
            else {
                max = (0, commons_1.$unsigned)(node._max);
            }
        }
        if (min > max) {
            errors.push(`maximal size < minimal size for array node`);
        }
        if ((0, commons_1.$ok)(itemParser.includedDefaultValue())) {
            errors.push(`_itemsParser cannot contain a default value`);
        }
        super((0, commons_1.$ok)(m) ? !!m : min > 0, node._checker, node._transformer, node._natifier);
        if ((0, commons_1.$defined)(m) && !(0, commons_1.$isbool)(m)) {
            this.errors.push(`_mandatory field should be a boolean`);
        }
        this._min = min;
        this._max = max;
        this._itemsParser = itemParser;
        for (let e of errors) {
            this.errors.push(e);
        }
    }
    nodeType() { return 'array'; }
    toJSON() {
        return {
            _type: 'array',
            _mandatory: this.mandatory,
            _min: this._min > 0 ? this._min : 'none',
            _max: this._max < types_1.UINT_MAX ? this._max : 'none',
            _check: this._check,
            _transform: this._transform,
            _natify: this._natify,
            _itemsStruct: this._itemsParser.toJSON()
        };
    }
    _validate(value, path, opts) {
        if (!super._validate(value, path, opts)) {
            return false;
        }
        if ((0, commons_1.$ok)(value) && !(0, commons_1.$isarray)(value)) {
            return _serror(opts, path, 'is not an array');
        }
        const n = (0, commons_1.$count)(value);
        if (n < this._min) {
            return _serror(opts, path, `.size is too small (${n} < ${this._min})`);
        }
        if (n > this._max) {
            return _serror(opts, path, `.size is too large (${n} > ${this._max})`);
        }
        const p = path.length ? path : 'array';
        let doomed = false;
        for (let i = 0; i < n; i++) {
            if (!this._itemsParser._validate(value[i], `${p}[${i}]`, opts)) {
                doomed = true;
            }
        }
        if (doomed) {
            return false;
        }
        return this._check(value, opts) ? true : _serror(opts, path, 'specific check did fail');
    }
    rawInterpret(value, opts) {
        if (!(0, commons_1.$ok)(value) && !this.mandatory) {
            return value;
        }
        const ret = [];
        for (let v of value) {
            ret.push(this._itemsParser.rawInterpret(v, opts));
        }
        return this._transform(ret, opts);
    }
    rawEncode(value, opts) {
        const ret = [];
        for (let v of value) {
            ret.push(this._itemsParser.rawEncode(v, opts));
        }
        return this._natify ? this._natify(ret, opts) : ret;
    }
}
function isLeafParserType(s) {
    return (0, commons_1.$length)(s) ? TSLeafParser.isLeafParserType(s) : false;
}
exports.isLeafParserType = isLeafParserType;
class TSLeafParser extends TSParser {
    static isLeafParserType(s) { return (0, commons_1.$ok)(TSLeafParser.__managers[s]); }
    constructor(node) {
        const manager = (0, commons_1.$isstring)(node._type) ? TSLeafParser.__managers[node._type] : undefined;
        const transformer = (0, commons_1.$ok)(node._transformer) ? node._transformer : ((0, commons_1.$ok)(manager === null || manager === void 0 ? void 0 : manager.trans) ? manager.trans : undefined);
        const natifier = (0, commons_1.$ok)(node._natifier) ? node._natifier : ((0, commons_1.$ok)(manager === null || manager === void 0 ? void 0 : manager.v2nat) ? manager.v2nat : undefined);
        const mandatory = !!node._mandatory;
        super(mandatory, node._checker, transformer, natifier);
        this._defaultValue = undefined;
        this._exportAsEnum = false;
        if (!(0, commons_1.$ok)(manager)) {
            this.errors.push(`Invalid parser type '${node._type}'`);
            this._manager = TSLeafParser.__dummyManager;
        }
        else {
            this._manager = manager;
        }
        const enumeration = node._enum;
        if ((0, commons_1.$isarray)(enumeration)) {
            if (!(0, commons_1.$ok)(manager === null || manager === void 0 ? void 0 : manager.enum)) {
                this.errors.push(`Parser type '${node._type}' does not support enumeration`);
            }
            if (!enumeration.length) {
                this.errors.push(`Empty enumeration definition array for type '${node._type}'`);
            }
            for (let e of enumeration) {
                if (!(manager === null || manager === void 0 ? void 0 : manager.enum(e))) {
                    this.errors.push(`Enumeration value '${e}' is invalid for type '${node._type}'`);
                }
            }
            this._enumeration = new Set(enumeration);
        }
        else if ((0, commons_1.$isobject)(enumeration)) {
            if (!(0, commons_1.$ok)(manager === null || manager === void 0 ? void 0 : manager.enum)) {
                this.errors.push(`Parser type '${node._type}' does not support enumeration`);
            }
            const entries = Object.entries(enumeration);
            this._enumeration = new Set();
            this._reverse = new Map();
            for (let [key, item] of entries) {
                if (!key.length || !key.match(IsFielOrEnumdRegex)) {
                    this.errors.push(`Wrong enumeration key '${key}' for type '${node._type}'`);
                }
                if (!!(manager === null || manager === void 0 ? void 0 : manager.enum(item))) {
                    this._enumeration.add(item);
                    this._reverse.set(item, key);
                }
                else {
                    this.errors.push(`Enumeration value '${item}' for key '${key}' is invalid for type '${node._type}'`);
                }
            }
            if (!this._enumeration.size) {
                this.errors.push(`Empty enumeration definition dictionary for type '${node._type}'`);
            }
            this._conversion = enumeration;
            this._exportAsEnum = !!node._exportAsEnum;
        }
        else if ((0, commons_1.$ok)(enumeration)) {
            this.errors.push(`Bad enumeration definition for type '${node._type}'`);
        }
        if ((0, commons_1.$ok)(node._options)) {
            this._options = node._options;
        }
        if ((0, commons_1.$ok)(node._default)) {
            if (mandatory) {
                this.errors.push(`Parser type '${node._type}' cannot be mandatory and have a default value at the same time`);
            }
            else if (!(manager === null || manager === void 0 ? void 0 : manager.valid(node._default))) {
                this.errors.push(`Parser type '${node._type}' cannot be have ${node._default} as default value`);
            }
            else {
                this._defaultValue = node._default;
            }
        }
        this._type = node._type;
    }
    get isKey() { return !!this._manager.iskey; }
    nodeType() { return this._type; }
    includedDefaultValue() { return this._defaultValue; }
    toJSON() {
        return {
            _type: this._type,
            _mandatory: this.mandatory,
            _enumeration: (0, commons_1.$ok)(this._enumeration) ? Array.from(this._enumeration) : 'none',
            _conversion: (0, commons_1.$value)(this._conversion, 'none'),
            _check: this._check,
            _transform: this._transform,
            _natify: this._natify,
            _defaultValue: this._defaultValue
        };
    }
    _localOptions(opts) {
        let localOptions = undefined;
        if ((0, commons_1.$ok)(opts)) {
            localOptions = (0, commons_1.$ok)(this._options) ? Object.assign(Object.assign({}, opts), { options: this._options }) : opts;
        }
        else if ((0, commons_1.$ok)(this._options)) {
            localOptions = { options: this._options };
        }
        return localOptions;
    }
    _validate(value, path, opts) {
        if (!(0, commons_1.$ok)(value)) {
            value = this._defaultValue;
        }
        if (!super._validate(value, path, opts)) {
            return false;
        }
        if (this._conversion && (0, commons_1.$isstring)(value)) {
            const tag = value;
            value = this._conversion[tag];
            if (!(0, commons_1.$ok)(value)) {
                return _serror(opts, path, `has not a valid enum tag ${tag}`);
            }
        }
        let lopts = this._localOptions(opts);
        if ((0, commons_1.$ok)(value) && !this._manager.valid(value, lopts)) {
            return _serror(opts, path, `is not a valid ${this._type}`);
        }
        if (this._enumeration) {
            if (!this._enumeration.has(this._convertIfString(value, lopts))) {
                const s = (0, commons_1.$isstring)(value) ? "'" + value + "'" : value;
                return _serror(opts, path, `has not a valid enum value ${s}`);
            }
        }
        return this._check(value, opts) ? true : _serror(opts, path, 'specific check did fail');
    }
    _convertIfString(value, opts) { return this._transform((0, commons_1.$isstring)(value) ? this._manager.str2v(value, opts) : value, opts); }
    rawInterpret(value, opts) {
        if (!(0, commons_1.$ok)(value)) {
            value = this._defaultValue;
        }
        if (!(0, commons_1.$ok)(value) && !this.mandatory) {
            return value;
        }
        if (this._conversion && (0, commons_1.$isstring)(value)) {
            value = this._conversion[value];
        }
        const r = this._convertIfString(value, this._localOptions(opts));
        return r;
    }
    rawEncode(value, opts) {
        if (this._conversion && (0, commons_1.$isstring)(value)) {
            value = (0, commons_1.$value)(this._conversion[value], value);
        }
        if (this._exportAsEnum && (0, commons_1.$ok)(this._reverse)) {
            value = (0, commons_1.$value)(this._reverse.get(value), value);
        }
        if (!(0, commons_1.$ok)(value)) {
            return value;
        }
        switch (typeof value) {
            case 'bigint': return Number(value);
            case 'boolean': return value;
            case 'number': return value;
            case 'object': return this._natify ? this._natify(value, this._localOptions(opts)) : (0, commons_1.$string)(value);
            case 'string': return value;
            case 'symbol': return (0, commons_1.$string)(value);
            default: return null;
        }
    }
}
TSLeafParser.__managers = {
    'boolean': { valid: _isBoolean, trans: $bool, str2v: $bool },
    'charset': { valid: _isCharset, str2v: (s) => tscharset_1.TSCharset.charset(s), v2nat: (v) => v.name },
    'color': { valid: _iscolor, str2v: (s) => tscolor_1.TSColor.fromString(s), v2nat: _color2str },
    'continent': { valid: _isContinent, str2v: (s) => (0, strings_1.$ftrim)(s).toUpperCase(), enum: _isContinent, iskey: true },
    'country': { valid: _isCountry, trans: _countryTrans, str2v: (s) => (0, strings_1.$ftrim)(s).toUpperCase(), v2nat: (v) => v.alpha2Code, enum: (v) => types_1.TSCountrySet.has(v), iskey: true },
    'currency': { valid: _isCurrency, str2v: (s) => (0, strings_1.$ftrim)(s).toUpperCase(), enum: _isCurrency, iskey: true },
    'data': { valid: _isData, str2v: _decodeb64, v2nat: _encodeb64 },
    'date': { valid: _isTsDate, str2v: (s) => tsdate_1.TSDate.fromIsoString(s), v2nat: (v) => v.toIsoString() },
    'email': { valid: (v) => (0, commons_1.$isemail)(v), str2v: (s) => (0, commons_1.$email)(s), iskey: true },
    'hexa': { valid: _isHexaData, str2v: _decodeHexa, v2nat: (v) => (0, data_1.$encodeHexa)(v) },
    'ipaddress': { valid: commons_1.$isipaddress, str2v: (s) => (0, strings_1.$ftrim)(s), enum: commons_1.$isipaddress, iskey: true },
    'ipv4': { valid: commons_1.$isipv4, str2v: (s) => (0, strings_1.$ftrim)(s), enum: commons_1.$isipv4, iskey: true },
    'ipv6': { valid: commons_1.$isipv6, str2v: (s) => (0, strings_1.$ftrim)(s), enum: commons_1.$isipv6, iskey: true },
    'int': { valid: (v) => _isInt(v, types_1.INT_MIN, types_1.INT_MAX), str2v: _int, enum: (v) => _isInt(v, types_1.INT_MIN, types_1.INT_MAX), iskey: true },
    'int8': { valid: (v) => _isInt(v, types_1.INT8_MIN, types_1.INT8_MAX), str2v: _int, enum: (v) => _isInt(v, types_1.INT8_MIN, types_1.INT8_MAX), iskey: true },
    'int16': { valid: (v) => _isInt(v, types_1.INT16_MIN, types_1.INT16_MAX), str2v: _int, enum: (v) => _isInt(v, types_1.INT16_MIN, types_1.INT16_MAX), iskey: true },
    'int32': { valid: (v) => _isInt(v, types_1.INT32_MIN, types_1.INT32_MAX), str2v: _int, enum: (v) => _isInt(v, types_1.INT32_MIN, types_1.INT32_MAX), iskey: true },
    'jsdate': { valid: _isJsDate, str2v: (s) => new Date(s), v2nat: (v) => v.toISOString() },
    'language': { valid: _isLanguage, str2v: (s) => (0, strings_1.$ftrim)(s).toLowerCase(), enum: _isLanguage, iskey: true },
    'native': { valid: _isNativeValue, str2v: (s) => s, v2nat: _nativeValue2nat },
    'number': { valid: _isNumber, str2v: (s) => Number(s), enum: (v) => (0, commons_1.$isnumber)(v) },
    'paper': { valid: _isDocumentFormat, str2v: (s) => (0, strings_1.$ftrim)(s).toLowerCase(), enum: _isDocumentFormat, iskey: true },
    'path': { valid: _isPath, str2v: (s) => s, enum: _isPath, iskey: true },
    'phone': { valid: (v) => (0, commons_1.$isphonenumber)(v), str2v: (s) => tsphonenumber_1.TSPhoneNumber.fromString(s), v2nat: (v) => v.standardNumber, iskey: true },
    'string': { valid: (v) => typeof v === 'string', str2v: (s) => s, enum: (v) => typeof v === 'string' && v.length > 0, iskey: true },
    'uint8': { valid: (v) => _isInt(v, 0, types_1.UINT8_MAX), str2v: _uint, enum: (v) => _isInt(v, 0, types_1.UINT8_MAX), iskey: true },
    'uint16': { valid: (v) => _isInt(v, 0, types_1.UINT16_MAX), str2v: _uint, enum: (v) => _isInt(v, 0, types_1.UINT16_MAX), iskey: true },
    'uint32': { valid: (v) => _isInt(v, 0, types_1.UINT32_MAX), str2v: _uint, enum: (v) => _isInt(v, 0, types_1.UINT32_MAX), iskey: true },
    'unsigned': { valid: (v) => _isInt(v, 0), str2v: _uint, enum: (v) => _isInt(v, 0), iskey: true },
    'url': { valid: _isURL, trans: _valueToTSURL, str2v: _stringToUrl, v2nat: (v) => v.href, enum: _isStringURL, iskey: true },
    'uuid': { valid: (v) => (0, commons_1.$isuuid)(v), str2v: (s) => (0, commons_1.$UUID)(s), enum: (v) => (0, commons_1.$isuuid)(v), iskey: true },
};
TSLeafParser.__dummyManager = { valid: (_) => false, str2v: (_) => undefined };
function $bool(v, opts) {
    switch (typeof v) {
        case 'bigint': {
            v = Number(v);
            return !isNaN(v) && v !== 0;
        }
        case 'boolean': return v;
        case 'number': return !isNaN(v) && v !== 0;
        case 'string': return !!_stringToBoolean(v, opts);
        case 'symbol': return !!_stringToBoolean((0, commons_1.$string)(v), opts);
        default: return false;
    }
}
exports.$bool = $bool;
function $validateParsedValue(v, type, opts) { var _a; return (0, commons_1.$ok)(v) && !!((_a = TSLeafParser.__managers[type]) === null || _a === void 0 ? void 0 : _a.valid(v, opts)); }
exports.$validateParsedValue = $validateParsedValue;
function _isNativeValue(v) {
    switch (typeof v) {
        case 'bigint':
            v = Number(v);
            return !isNaN(v) && isFinite(v);
        case 'number': return !isNaN(v) && isFinite(v);
        case 'boolean':
        case 'string':
        case 'symbol': return true;
        default: return false;
    }
}
function _nativeValue2nat(v) {
    switch (typeof v) {
        case 'bigint': return Number(v);
        case 'number':
        case 'boolean':
        case 'string': return v;
        case 'symbol': return (0, commons_1.$string)(v);
        case 'undefined': return undefined;
        default: return null;
    }
}
function _isCharset(v) { return v instanceof tscharset_1.TSCharset || ((0, commons_1.$isstring)(v) && (0, commons_1.$ok)(tscharset_1.TSCharset.charset(v))); }
function _isContinent(v) { return (0, commons_1.$isstring)(v) && types_1.TSContinentSet.has((0, strings_1.$ftrim)(v).toUpperCase()); }
function _isCurrency(v) { return (0, commons_1.$isstring)(v) && types_1.TSCurrencySet.has((0, strings_1.$ftrim)(v).toUpperCase()); }
function _isDocumentFormat(v) { return (0, commons_1.$isstring)(v) && (0, commons_1.$ok)(tsgeometry_1.TSDocumentFormats[(0, strings_1.$ftrim)(v).toLowerCase()]); }
function _isLanguage(v) { return (0, commons_1.$isstring)(v) && types_1.TSLanguageSet.has((0, strings_1.$ftrim)(v).toLowerCase()); }
function _isNumber(v) { return (0, commons_1.$isnumber)(v) || ((0, commons_1.$isstring)(v) && (0, commons_1.$isnumber)(Number(v))); }
function _isCountry(v) {
    if (v instanceof tscountry_1.TSCountry) {
        return true;
    }
    if ((0, commons_1.$isstring)(v)) {
        const s = (0, strings_1.$ftrim)(v).toUpperCase();
        return (s.length === 2 && types_1.TSCountrySet.has(s)) || (s.length === 3 && (0, commons_1.$ok)(tscountry_1.TSCountry.alpha2CodeForAlpha3Code(s)));
    }
    return false;
}
function _countryTrans(v) {
    if (v instanceof tscountry_1.TSCountry) {
        return v.alpha2Code;
    }
    return v.length === 3 ? tscountry_1.TSCountry.alpha2CodeForAlpha3Code(v) : v;
}
function _stringToBoolean(s, opts) {
    s = (0, strings_1.$ascii)((0, strings_1.$ftrim)(s)).toLowerCase();
    if ((opts === null || opts === void 0 ? void 0 : opts.context) === TSParserActionContext.json) {
        return s === 'true' ? true : (s === 'false' ? false : null);
    }
    return s === 'true' || s === '1' || s === 'y' || s === 'yes' ? true : (s === 'false' || s === '0' || s === 'n' || s === 'no' ? false : null);
}
function _isBoolean(v, opts) { return (0, commons_1.$isbool)(v) || v === 0 || v === 1 || ((0, commons_1.$isstring)(v) && (0, commons_1.$ok)(_stringToBoolean(v, opts))); }
function _isPath(v, opts) {
    var _a;
    if (typeof v === 'string' && (0, strings_1.$ftrim)(v).length > 0) {
        if (v.toLowerCase().startsWith('file:')) {
            return false;
        }
        const absolute = v.startsWith('/');
        return (!((_a = opts === null || opts === void 0 ? void 0 : opts.options) === null || _a === void 0 ? void 0 : _a.absolute) || absolute) && (0, commons_1.$isurl)('file://localhost' + (absolute ? v : '/' + v));
    }
    return false;
}
function _isURL(v, opts) { var _a, _b; return (0, commons_1.$isurl)(v, { refusesParameters: (opts === null || opts === void 0 ? void 0 : opts.context) === TSParserActionContext.url || !!((_a = opts === null || opts === void 0 ? void 0 : opts.options) === null || _a === void 0 ? void 0 : _a.refusesParameters), acceptedProtocols: (_b = opts === null || opts === void 0 ? void 0 : opts.options) === null || _b === void 0 ? void 0 : _b.acceptedProtocols }); }
function _isStringURL(v, opts) { return typeof v === 'string' && _isURL(v, opts); }
function _stringToUrl(s, opts) { var _a, _b; return tsurl_1.TSURL.url(s, { refusesParameters: (opts === null || opts === void 0 ? void 0 : opts.context) === TSParserActionContext.url || !!((_a = opts === null || opts === void 0 ? void 0 : opts.options) === null || _a === void 0 ? void 0 : _a.refusesParameters), acceptedProtocols: (_b = opts === null || opts === void 0 ? void 0 : opts.options) === null || _b === void 0 ? void 0 : _b.acceptedProtocols }); }
function _valueToTSURL(v, opts) {
    var _a, _b;
    if (v instanceof tsurl_1.TSURL) {
        return v;
    }
    if (!(v instanceof URL)) {
        tserrors_1.TSError.throw('Impossible to convertany object to TSURL');
    }
    ;
    const ret = tsurl_1.TSURL.from(v, { refusesParameters: (opts === null || opts === void 0 ? void 0 : opts.context) === TSParserActionContext.url || !!((_a = opts === null || opts === void 0 ? void 0 : opts.options) === null || _a === void 0 ? void 0 : _a.refusesParameters), acceptedProtocols: (_b = opts === null || opts === void 0 ? void 0 : opts.options) === null || _b === void 0 ? void 0 : _b.acceptedProtocols });
    if (!(0, commons_1.$ok)(ret)) {
        tserrors_1.TSError.throw('Impossible to convert URL to TSURL');
    }
    ;
    return ret;
}
const _b64regex = /^[A-Za-z0-9\+\/]+[\=]*$/;
const _b64URLregex = /^[A-Za-z0-9\-\_]+[\=]*$/;
function _isData(v, opts) { return (0, commons_1.$isdataobject)(v) || ((0, commons_1.$isstring)(v) && ((opts === null || opts === void 0 ? void 0 : opts.context) === TSParserActionContext.url ? _b64URLregex : _b64regex).test(v)); }
function _encodeb64(v, opts) { return (opts === null || opts === void 0 ? void 0 : opts.context) === TSParserActionContext.url ? (0, data_1.$encodeBase64URL)(v) : (0, data_1.$encodeBase64)(v); }
function _decodeb64(s, opts) { return (opts === null || opts === void 0 ? void 0 : opts.context) === TSParserActionContext.url ? (0, data_1.$decodeBase64URL)(s) : (0, data_1.$decodeBase64)(s); }
function _decodeHexa(s) { return Buffer.from(s, 'hex'); }
const _hexaRegex = /^[A-Fa-f0-9]+$/;
function _isHexaData(v) { return (0, commons_1.$isdataobject)(v) || ((0, commons_1.$isstring)(v) && _hexaRegex.test(v)); }
function _int(v) { return (0, commons_1.$int)(v); }
function _uint(v) { return (0, commons_1.$unsigned)(v); }
function _isInt(v, min = types_1.INT_MIN, max = types_1.INT_MAX) { return (0, commons_1.$isint)(v, min, max) || ((0, commons_1.$isstring)(v) && (0, commons_1.$isint)(Number(v), min, max)); }
function _isJsDate(v) { return v instanceof Date || ((0, commons_1.$isstring)(v) && v !== '0' && (0, commons_1.$isnumber)(Date.parse(v))); }
function _isTsDate(v) { return v instanceof tsdate_1.TSDate || ((0, commons_1.$isstring)(v) && (0, commons_1.$ok)((0, commons_1.$isodate)(v))); }
function _iscolor(o) { return o instanceof tscolor_1.TSColor || ((0, commons_1.$isstring)(o)) && (0, commons_1.$ok)(tscolor_1.TSColor.fromString(o)); }
function _color2str(v) { return v.toString({ colorSpace: tscolor_1.TSColorSpace.RGB, removeAlpha: false, rgbaCSSLike: true, shortestCSS: false, uppercase: false }); }
function _serror(opts, path, error) {
    var _a;
    (_a = opts === null || opts === void 0 ? void 0 : opts.errors) === null || _a === void 0 ? void 0 : _a.push(`${path.length > 0 ? path : 'value'}${error.startsWith('.') ? '' : ' '}${error}${error.endsWith('') ? '' : '.'}`);
    return false;
}
const IsFielOrEnumdRegex = /^[a-zA-Z][a-zA-Z0-9_-]*$/g;
const InternalFieldsSet = new Set(['_mandatory', '_default', '_keysCase', '_checker', '_transformer', '_natifier', '_acceptsUncheckedItems', '_aliases', '_outputAliases', '_aliasUnsensitive']);
const InternalCasingMap = {
    'standard': (s) => s,
    'lowercase': (s) => s.toLowerCase(),
    'uppercase': (s) => s.toUpperCase()
};
function _structureConstruction(node, errors) {
    if ((0, commons_1.$isstring)(node)) {
        let mandatory = node.endsWith('!');
        if (mandatory) {
            node = node.left(node.length - 1);
        }
        return new TSLeafParser({ _type: node, _mandatory: mandatory });
    }
    else if ((0, commons_1.$isarray)(node)) {
        const a = node;
        if (a.length === 0 || a.length > 3) {
            errors === null || errors === void 0 ? void 0 : errors.push('wrong empty array node definition');
            return null;
        }
        let min = undefined, max = undefined;
        if (a.length > 1 && (0, commons_1.$ok)(a[1])) {
            if (!(0, commons_1.$isunsigned)(a[1])) {
                errors === null || errors === void 0 ? void 0 : errors.push(`wrong minimal size definition for array node`);
            }
            else {
                min = (0, commons_1.$unsigned)(a[1]);
            }
        }
        if (a.length > 2 && (0, commons_1.$ok)(a[2])) {
            if (!(0, commons_1.$isunsigned)(a[2])) {
                errors === null || errors === void 0 ? void 0 : errors.push(`wrong maximal size definition for array node`);
            }
            else {
                max = (0, commons_1.$unsigned)(a[2]);
            }
        }
        const itemsType = a[0];
        const itemsParser = _structureConstruction(itemsType, errors);
        if (!(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.isValid)) {
            if ((0, commons_1.$count)(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.errors)) {
                errors === null || errors === void 0 ? void 0 : errors.push(...itemsParser.errors);
            }
            return null;
        }
        return new TSArrayParser({ _itemsType: itemsType, _min: min, _max: max }, itemsParser);
    }
    else if ((0, commons_1.$isobject)(node)) {
        if ((0, commons_1.$ok)(node._type)) {
            return new TSLeafParser(node);
        }
        else if ((0, commons_1.$ok)(node._keysType) || (0, commons_1.$ok)(node._valueItemsType)) {
            if ((0, commons_1.$ok)(node._acceptsUncheckedItems)) {
                errors === null || errors === void 0 ? void 0 : errors.push('_acceptsUncheckedItems field cannot be set for a dictionary node definition');
            }
            if (!(0, commons_1.$ok)(node._keysType)) {
                errors === null || errors === void 0 ? void 0 : errors.push('_keysType field is not set for dictionary node definition');
            }
            const keysParser = _structureConstruction(node._keysType, errors);
            if (!(keysParser === null || keysParser === void 0 ? void 0 : keysParser.isValid)) {
                if ((0, commons_1.$count)(keysParser === null || keysParser === void 0 ? void 0 : keysParser.errors)) {
                    errors === null || errors === void 0 ? void 0 : errors.push(...keysParser.errors);
                }
                return null;
            }
            if (!(0, commons_1.$ok)(node._valueItemsType)) {
                errors === null || errors === void 0 ? void 0 : errors.push('_valueItemsType field is not set for a dictionary node definition');
            }
            const itemsParser = _structureConstruction(node._valueItemsType, errors);
            if (!(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.isValid)) {
                if ((0, commons_1.$count)(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.errors)) {
                    errors === null || errors === void 0 ? void 0 : errors.push(...itemsParser.errors);
                }
                return null;
            }
            return new TSDictionaryParser(node, keysParser, itemsParser);
        }
        else if ((0, commons_1.$ok)(node._itemsType)) {
            const itemsParser = _structureConstruction(node._itemsType, errors);
            if (!(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.isValid)) {
                if ((0, commons_1.$count)(itemsParser === null || itemsParser === void 0 ? void 0 : itemsParser.errors)) {
                    errors === null || errors === void 0 ? void 0 : errors.push(...itemsParser.errors);
                }
                return null;
            }
            return new TSArrayParser(node, itemsParser);
        }
        else {
            const itemsParsers = {};
            const entries = Object.entries(node);
            const keyscase = (0, commons_1.$value)(InternalCasingMap[(0, commons_1.$value)(node._keysCase, TSCase.standard)], (s) => s);
            let found = 0;
            for (let [key, subnode] of entries) {
                if (!InternalFieldsSet.has(key)) {
                    if (!key.match(IsFielOrEnumdRegex)) {
                        errors === null || errors === void 0 ? void 0 : errors.push(`wrong key '${key}' for object node definition`);
                    }
                    else {
                        const itemParser = _structureConstruction(subnode, errors);
                        if (!(itemParser === null || itemParser === void 0 ? void 0 : itemParser.isValid)) {
                            if ((0, commons_1.$count)(itemParser === null || itemParser === void 0 ? void 0 : itemParser.errors)) {
                                errors === null || errors === void 0 ? void 0 : errors.push(...itemParser.errors);
                            }
                            return null;
                        }
                        itemsParsers[keyscase(key)] = itemParser;
                        found++;
                    }
                }
            }
            if (!found) {
                errors === null || errors === void 0 ? void 0 : errors.push('empty object node definition');
                return null;
            }
            return new TSObjectParser(node, itemsParsers, !!(node._acceptsUncheckedItems));
        }
    }
    return null;
}
//# sourceMappingURL=tsparser.js.map