"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSPhoneNumber = exports.PhoneValidity = void 0;
const commons_1 = require("./commons");
const strings_1 = require("./strings");
const tscountry_1 = require("./tscountry");
const tsdefaults_1 = require("./tsdefaults");
const types_1 = require("./types");
const customInspectSymbol = Symbol.for('nodejs.util.inspect.custom');
var PhoneValidity;
(function (PhoneValidity) {
    PhoneValidity[PhoneValidity["OK"] = 0] = "OK";
    PhoneValidity[PhoneValidity["WrongLength"] = 1] = "WrongLength";
    PhoneValidity[PhoneValidity["MalformedNumber"] = 2] = "MalformedNumber";
    PhoneValidity[PhoneValidity["BadCountryNumber"] = 3] = "BadCountryNumber";
    PhoneValidity[PhoneValidity["CountryNotFound"] = 4] = "CountryNotFound";
    PhoneValidity[PhoneValidity["MixedCountries"] = 5] = "MixedCountries";
})(PhoneValidity = exports.PhoneValidity || (exports.PhoneValidity = {}));
;
class TSPhoneNumber {
    static fromString(source, telcountry) {
        const [phoneNumber, phoneCountry, validity] = _phoneFromString(source, telcountry);
        return validity === PhoneValidity.OK ? new TSPhoneNumber(phoneNumber, phoneCountry) : null;
    }
    static interpret(source, telcountry) {
        const [phoneNumber, phoneCountry, validity] = _phoneFromString(source, telcountry);
        return validity === PhoneValidity.OK ? new TSPhoneNumber(phoneNumber, phoneCountry) : validity;
    }
    static validity(source, telcountry) {
        const [, , validity] = _phoneFromString(source, telcountry);
        return validity;
    }
    constructor(number, country) {
        this.number = number;
        this.country = country;
        this.leafInspect = TSPhoneNumber.prototype._standardNumber;
        this.toString = TSPhoneNumber.prototype._standardNumber;
        this.toJSON = TSPhoneNumber.prototype._standardNumber;
    }
    get dialCode() { return this.country.phonePlan.dialCode; }
    get standardNumber() { return this._standardNumber(); }
    clone() { return this; }
    [customInspectSymbol](depth, inspectOptions, inspect) {
        return this.leafInspect();
    }
    compare(other) {
        if (this === other) {
            return types_1.Same;
        }
        if ((0, commons_1.$isstring)(other)) {
            other = TSPhoneNumber.fromString(other);
        }
        if (!(other instanceof TSPhoneNumber)) {
            return undefined;
        }
        const a = this._standardNumber();
        const b = other._standardNumber();
        return a > b ? types_1.Descending : (a < b ? types_1.Ascending : types_1.Same);
    }
    isEqual(other) {
        if (this === other) {
            return true;
        }
        if ((0, commons_1.$isstring)(other)) {
            other = TSPhoneNumber.fromString(other);
        }
        return other instanceof TSPhoneNumber ? this._standardNumber() === other._standardNumber() : false;
    }
    toArray() { return [this]; }
    _standardNumber() {
        const p = this.country.phonePlan;
        const t = p.trunkCode;
        const n = this.number;
        return `+${p.dialCode} ${t.length > 0 && n.startsWith(t) ? n.slice(t.length) : n}`;
    }
}
exports.TSPhoneNumber = TSPhoneNumber;
function _phoneFromString(source, telcountry) {
    let s = (0, strings_1.$ascii)((0, strings_1.$normspaces)(source, { replacer: "", strict: true }));
    let len = s.length;
    let min = 5;
    if (s.startsWith('(')) {
        if (s.slice(1, 2) === '+') {
            s = '+(' + s.slice(2);
        }
        else if (s.slice(1, 3) === '00') {
            s = s.slice(3);
            if (s.length < 2) {
                return [null, null, PhoneValidity.MalformedNumber];
            }
            ;
            s = '+(' + (_isPhoneSeparator(s[0]) ? s.slice(1) : s);
        }
        len = s.length;
        min += 2;
    }
    if (len < min) {
        return [null, null, PhoneValidity.WrongLength];
    }
    let start = 0;
    let dialCode = '';
    let countries = [];
    let localNumber = false;
    if (s[0] === '+' || s.startsWith('00')) {
        start = s[0] === '+' ? 1 : 2;
        const ds = _removeSeparators(s.slice(start));
        if (!(0, commons_1.$ok)(ds)) {
            return [null, null, PhoneValidity.MalformedNumber];
        }
        s = ds;
        len = s.length;
        if (len < 5) {
            return [null, null, PhoneValidity.WrongLength];
        }
        [dialCode, countries] = _findDialCode(s);
        start = dialCode.length;
        if (!start || !countries.length) {
            return [null, null, PhoneValidity.CountryNotFound];
        }
        if ((0, commons_1.$ok)(telcountry) && !countries.includes(telcountry)) {
            return [null, null, PhoneValidity.BadCountryNumber];
        }
        else if ((0, commons_1.$ok)(telcountry)) {
            countries = [telcountry];
        }
    }
    else {
        const defaultCountry = (0, commons_1.$value)(telcountry, tsdefaults_1.TSDefaults.defaults().defaultCountry);
        dialCode = defaultCountry.phonePlan.dialCode;
        countries = [defaultCountry];
        localNumber = true;
    }
    s = s.slice(start);
    len = s.length;
    countries = countries.filter(c => {
        const t = c.phonePlan.trunkCode;
        const min = c.phonePlan.minDigits - (!localNumber && t.length > 0 && !s.startsWith(t) ? t.length : 0);
        return len >= min && len <= c.phonePlan.maxDigits;
    });
    if (!countries.length) {
        return [null, null, PhoneValidity.WrongLength];
    }
    const countriesWithValidAreaCodes = countries.filter(c => _numberInAreaCodes(s, c));
    countries = countriesWithValidAreaCodes.length > 0 ? countriesWithValidAreaCodes : countries.filter(c => c.phonePlan.areaCodes.length === 0);
    if (!countries.length) {
        return [null, null, PhoneValidity.CountryNotFound];
    }
    const c = countries[0];
    const t = c.phonePlan.trunkCode;
    if (!localNumber && t.length > 0 && !s.startsWith(t)) {
        s = t + s;
    }
    if (countries.length > 1) {
        return [s, null, PhoneValidity.MixedCountries];
    }
    if ((0, commons_1.$ok)(telcountry) && c !== telcountry) {
        return [s, c, PhoneValidity.BadCountryNumber];
    }
    return [s, c, PhoneValidity.OK];
}
function _numberInAreaCodes(s, c) {
    const areaCodes = c.phonePlan.areaCodes;
    if (areaCodes.length) {
        for (let ac of areaCodes) {
            if (s.startsWith(ac))
                return true;
        }
    }
    return false;
}
function _removeSeparators(s) {
    let ret = '';
    let foundOpenParenthesis = false;
    let foundClosedParenthesis = false;
    let SepState;
    (function (SepState) {
        SepState[SepState["Start"] = 0] = "Start";
        SepState[SepState["None"] = 1] = "None";
        SepState[SepState["Last"] = 2] = "Last";
    })(SepState || (SepState = {}));
    ;
    let sepstate = SepState.Start;
    for (let c of s) {
        if (c === '(') {
            if (sepstate === SepState.Last || foundOpenParenthesis) {
                return null;
            }
            foundOpenParenthesis = true;
            sepstate = SepState.Last;
        }
        else if (c === ')') {
            if (sepstate !== SepState.None || !foundOpenParenthesis || foundClosedParenthesis) {
                return null;
            }
            foundClosedParenthesis = true;
            sepstate = SepState.None;
        }
        else if (c >= '0' && c <= '9') {
            ret += c;
            sepstate = SepState.None;
        }
        else if (_isPhoneSeparator(c)) {
            if (sepstate !== SepState.None) {
                return null;
            }
            sepstate = SepState.Last;
        }
        else {
            return null;
        }
    }
    return sepstate === SepState.None && (!foundOpenParenthesis || foundClosedParenthesis) ? ret : null;
}
function _isPhoneSeparator(c) { return c === '.' || c === '-' || c === '~' || c === '/'; }
function _findDialCode(s) {
    for (let i = 1; i < 4; i++) {
        const dialCode = s.slice(0, i);
        const countries = tscountry_1.TSCountry.countriesForDialCode(dialCode);
        if (countries.length > 0) {
            return [dialCode, countries];
        }
    }
    return ['', []];
}
//# sourceMappingURL=tsphonenumber.js.map