"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSStaticWebsite = void 0;
const commons_1 = require("./commons");
const fs_1 = require("./fs");
const tserrors_1 = require("./tserrors");
const types_1 = require("./types");
const utils_1 = require("./utils");
class TSStaticWebsite {
    constructor(uri, definition) {
        const def = (0, commons_1.$isstring)(definition) ? { folder: definition } : definition;
        if (!(0, commons_1.$length)(uri)) {
            tserrors_1.TSError.throw(`TSStaticWebsite.constructor(): url not defined for folder '${def.folder}'`, Object.assign({ uri: uri }, def));
        }
        if (!(0, fs_1.$isdirectory)(def.folder)) {
            tserrors_1.TSError.throw(`TSStaticWebsite.constructor(): website ${uri} root folder ${def.folder} was not found on disk.`, Object.assign({ uri: uri }, def));
        }
        this.uri = uri.toLowerCase();
        this.folder = def.folder;
        this._maxCacheSize = _uint32v(def.maxCacheSize, 128 * 1024 * 1024);
        this._maxCachedFiles = _uint32v(def.maxCachedFiles, 10000);
        this._maxCachedFileSize = _uint32v(def.maxCachedFileSize, 256 * 1024);
        this._maxBlacklistedFiles = _uint32v(def.maxBlacklistedFiles, 50000);
        this._cache = new Map();
        this._cacheMemory = 0;
        this._blacklisted = new Set();
        this._mimeChecker = (0, commons_1.$value)(def.mimeChecker, (s) => {
            return (0, commons_1.$length)(s) ? TSStaticWebsite.__TSServerStandardsTypes[s] : s;
        });
    }
    getStaticResource(uri) {
        let ret = undefined;
        let type = '';
        const lcUri = uri.toLowerCase();
        if (lcUri.startsWith(this.uri)) {
            const ext = (0, fs_1.$ext)(lcUri);
            type = (0, commons_1.$value)(this._mimeChecker(ext), '');
            if (type.length) {
                const path = uri.slice(this.uri.length);
                if (!this._blacklisted.has(path)) {
                    const cache = this._cache.get(path);
                    if ((0, commons_1.$ok)(cache)) {
                        ret = cache.buffer;
                        cache.usage++;
                    }
                    else {
                        const file = (0, fs_1.$path)(this.folder, path);
                        if ((0, fs_1.$isfile)(file)) {
                            ret = (0, fs_1.$readBuffer)(file);
                            if ((0, commons_1.$ok)(ret)) {
                                this._cacheContent(path, ret);
                            }
                            else {
                                this._blacklist(path);
                            }
                        }
                        else {
                            this._blacklist(path);
                        }
                    }
                }
            }
        }
        return [ret, type];
    }
    _cacheContent(path, b) {
        if (this._cache.size < this._maxCachedFiles && b.length <= this._maxCachedFileSize && this._cacheMemory + b.length <= this._maxCacheSize) {
            this._cache.set(path, { buffer: b, timestamp: (0, utils_1.$mark)(), usage: 1 });
            this._cacheMemory += b.length;
        }
    }
    _blacklist(path) {
        if (this._blacklisted.size < this._maxBlacklistedFiles) {
            this._blacklisted.add(path);
        }
    }
    clearCaches() {
        this._cacheMemory = 0;
        if (this._blacklisted.size > this._maxBlacklistedFiles * 0.75) {
            this._blacklisted.clear();
        }
        this._cache.conditionalClear((_, v) => {
            if (!v.usage) {
                return true;
            }
            v.usage = 0;
            this._cacheMemory += v.buffer.length;
            return false;
        });
    }
}
exports.TSStaticWebsite = TSStaticWebsite;
TSStaticWebsite.__TSServerStandardsTypes = {
    'html': 'text/html',
    'htm': 'text/html',
    'css': 'text/css',
    'js': 'text/javascript',
    'svg': 'image/svg+xml',
    'svgz': 'image/svg+xml',
    'jpeg': 'image/jpeg',
    'jpg': 'image/jpg',
    'png': 'image/png',
    'gif': 'image/gif',
    'ico': 'image/x-icon',
    'pdf': 'application/pdf'
};
function _uint32v(v, defaultValue) {
    return ((0, commons_1.$ok)(v) && (0, commons_1.$isunsigned)(v, types_1.UINT32_MAX) ? v : defaultValue);
}
//# sourceMappingURL=tsserver_websites.js.map