"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSURL = void 0;
const types_1 = require("./types");
const commons_1 = require("./commons");
const strings_1 = require("./strings");
const tserrors_1 = require("./tserrors");
class TSURL {
    static from(source, options) {
        if (!(0, commons_1.$ok)(options)) {
            options = {};
            const protocol = source.protocol.toLowerCase();
            const w3c = TSURL.StandardURLProtocols.has(source.protocol.toLowerCase());
            if (!w3c && protocol.length) {
                options.acceptedProtocols = [protocol];
            }
            if (!source.search.length) {
                options.refusesParameters = true;
            }
        }
        return TSURL.url(source.href, options);
    }
    static compose(origin, path, options) {
        let o = (0, strings_1.$ftrim)(origin);
        let p = (0, strings_1.$ftrim)(path);
        if (!o.length) {
            o = 'http://localhost/';
        }
        if (!o.endsWith('/')) {
            o += '/';
        }
        if (p.startsWith('/')) {
            p = p.slice(1);
        }
        return this.url(o + p, options);
    }
    static url(source, options) {
        const opts = (0, commons_1.$value)(options, {});
        const url = _normalizeURLString(source);
        const len = url.length;
        const [protocol, w3c, aps] = _validateProtocolInfo(_findProtocol(url), opts.acceptedProtocols);
        let p = protocol.length;
        if (!p) {
            if (!!opts.throwsError) {
                new tserrors_1.TSError(`TSURL: bad protocol in '${url}'`);
            }
            return null;
        }
        if (url[p] !== '/' || url[p + 1] !== '/') {
            if (!!opts.throwsError) {
                new tserrors_1.TSError(`TSURL: unable to find // before hostname in url '${url}'`);
            }
            return null;
        }
        let c = '';
        p += 2;
        let hostendpos = p;
        while (hostendpos < len) {
            c = url[hostendpos];
            if (c === '?' || c === '#' || c === '/') {
                break;
            }
            hostendpos++;
        }
        const potentialHost = url.slice(p, hostendpos);
        const atpos = potentialHost.lastIndexOf('@');
        let auth = '';
        if (atpos >= 0) {
            auth = decodeURIComponent(url.slice(p, p + atpos));
            p += atpos + 1;
        }
        hostendpos = p;
        while (hostendpos < len) {
            if (TSURL.EndhostCharSet.has(url[hostendpos])) {
                break;
            }
            hostendpos++;
        }
        let [hostname, port] = _parseHostAndPort(url.slice(p, hostendpos));
        let path = '';
        if (!hostname.length || hostname.length > TSURL.HostMaxLength) {
            if (!!opts.throwsError) {
                new tserrors_1.TSError(`TSURL: bad hostname in url '${url}'`);
            }
            return null;
        }
        let ipv6Hostname = hostname.startsWith('[') && hostname.endsWith(']');
        if (ipv6Hostname) {
            hostname = hostname.slice(1, hostname.length - 1);
            if (!(0, commons_1.$isipv6)(hostname)) {
                if (!!opts.throwsError) {
                    new tserrors_1.TSError(`TSURL: bad ipv6 hostname in url '${url}'`);
                }
                return null;
            }
        }
        else {
            if (!_checkPotentialIPV4HostName(hostname)) {
                if (!!opts.throwsError) {
                    new tserrors_1.TSError(`TSURL: bad ipv4 hostname in url '${url}'`);
                }
                return null;
            }
            [hostname, path] = _validateHostName(hostname);
        }
        p = hostendpos;
        for (let p = hostendpos; p < len; p++) {
            const c = url[p];
            if (c === "'") {
                path += '%27';
            }
            else {
                path += TSURL.AutoEscapeCharSet.has(c) ? encodeURIComponent(c) : c;
            }
        }
        let hash = '';
        const hashIndex = path.indexOf('#');
        if (hashIndex >= 0) {
            hash = path.slice(hashIndex);
            path = path.slice(0, hashIndex);
        }
        let search = '';
        const qmIndex = path.indexOf('?');
        if (qmIndex >= 0) {
            search = path.slice(qmIndex);
            if (search.length === 1) {
                search = '';
            }
            path = path.slice(0, qmIndex);
        }
        if (search.length > 0 && !!opts.refusesParameters) {
            if (!!opts.throwsError) {
                new tserrors_1.TSError(`TSURL: unexpected parameters in url '${url}'`);
            }
            return null;
        }
        if (!path.length) {
            path = '/';
        }
        return new TSURL(protocol, hostname, port, path, hash, auth, search, ipv6Hostname, w3c, aps);
    }
    constructor(protocol, hostname, port, pathname, hash, auth, search, ipv6, w3c, protocols) {
        this._searchParams = undefined;
        this._href = undefined;
        this.toJSON = TSURL.prototype.toString;
        this._auth = auth;
        this._hash = hash;
        this._hostname = hostname;
        this._pathname = pathname;
        this._port = port;
        this._protocol = protocol;
        this._search = search;
        this._w3cProtocol = w3c;
        this._ipv6 = ipv6;
        this._acceptedProtocols = protocols;
    }
    get href() {
        if (!(0, commons_1.$defined)(this._href)) {
            this._href = this._calculated_href();
        }
        return this._href;
    }
    get w3c() { return this._w3cProtocol; }
    get origin() { return this._protocol + '//' + this.host; }
    get host() {
        let ret = this._ipv6 ? '[' + this._hostname + ']' : this._hostname;
        if (this._port.length) {
            ret += ':';
            ret += this._port;
        }
        return ret;
    }
    get searchParams() {
        if (!(0, commons_1.$defined)(this._searchParams)) {
            this._searchParams = new URLSearchParams(this.search);
        }
        return this._searchParams;
    }
    get auth() { return this._auth; }
    get hash() { return this._hash; }
    get search() { return this._search; }
    get pathname() { return this._pathname; }
    set pathname(s) {
        let path = '';
        for (const c of s) {
            if (c === "'") {
                path += '%27';
            }
            else {
                path += TSURL.AutoEscapeCharSet.has(c) ? encodeURIComponent(c) : c;
            }
        }
        if (!path.length) {
            path = '/';
        }
        this.pathname = path;
        this._href = undefined;
    }
    get hostname() { return this._hostname; }
    get port() { return this._port; }
    set port(p) { this._port = p; this._href = undefined; }
    get protocol() { return this._protocol; }
    set protocol(s) {
        const [protocol, w3c,] = _validateProtocolInfo(s, this._acceptedProtocols);
        if (!protocol.length) {
            tserrors_1.TSError.throw(`TSURL.protocol = value. Impossible to set new protocol '${s}'.`);
        }
        this._w3cProtocol = w3c;
        this._protocol = protocol;
        this._href = undefined;
    }
    isEqual(other) {
        return this === other || (other instanceof TSURL && this.href === other.href);
    }
    compare(other) {
        if (other === this) {
            return types_1.Same;
        }
        if (!(other instanceof TSURL)) {
            return undefined;
        }
        return this.href < other.href ? types_1.Ascending : (this.href > other.href ? types_1.Descending : types_1.Same);
    }
    toString() { return this.href; }
    toArray() { return [this]; }
    clone() {
        return new TSURL(this._protocol, this._hostname, this._port, this._pathname, this._hash, this._auth, this._search, this._ipv6, this._w3cProtocol, this._acceptedProtocols);
    }
    _calculated_href() {
        let auth = this._auth;
        if (auth.length) {
            auth = encodeURIComponent(auth);
            auth = auth.replace(/%3A/i, ':');
            auth += '@';
        }
        let hash = this._hash;
        if (hash.length && !hash.startsWith('#')) {
            hash = '#' + hash;
        }
        let search = this._search;
        if (search.length) {
            search = search.replace('#', '%23');
            if (!search.startsWith('?')) {
                search = '?' + search;
            }
        }
        let pathname = this._pathname;
        if (!pathname.startsWith('/')) {
            pathname = '/' + pathname;
        }
        return (this._protocol + '//' + auth + this.host + pathname + search + hash);
    }
}
exports.TSURL = TSURL;
TSURL.HostMaxLength = 255;
TSURL.AutoEscapeChars = ['{', '}', '|', '\\', '^', '`', "'", '<', '>', '"', '`', ' ', '\r', '\n', '\t'];
TSURL.StandardURLProtocols = new Set(['http:', 'https:', 'file:', 'ftp:', 'ws:', 'wss:']);
TSURL.AutoEscapeCharSet = new Set(TSURL.AutoEscapeChars);
TSURL.EndhostCharSet = new Set(['%', '/', '?', ';', '#', ...TSURL.AutoEscapeChars]);
const __protocolRegex = /^([a-z0-9.+-]+:)/i;
const __protocolCompleteRegex = /^([a-z0-9.+-]+:)$/i;
const __backslashRegex = /\\/g;
const __portRegex = /:[0-9]*$/;
const __mayBeIPV4Regex = /^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/;
const __hostnamePartRegex = /^[+a-z0-9A-Z_-]{0,63}$/;
const __hostnamePartStartRegex = /^([+a-z0-9A-Z_-]{0,63})(.*)$/;
const __nonASCIIGlobalRegex = /[^\x00-\x7F]/g;
function _checkPotentialIPV4HostName(hostname) {
    return !__mayBeIPV4Regex.test(hostname) || (0, commons_1.$isipv4)(hostname);
}
function _normalizeURLString(source) {
    let s = (0, strings_1.$ftrim)(source);
    if (s.startsWith('file:///')) {
        s = 'file://localhost/' + s.slice(8);
    }
    const index = s.indexOf('?');
    const splitchar = index > -1 && index < s.indexOf('#') ? '?' : '#';
    const parts = s.split(splitchar);
    parts[0] = parts[0].replace(__backslashRegex, '/');
    return parts.join(splitchar);
}
function _validateProtocolInfo(s, acceptedProtocols) {
    const aps = (0, commons_1.$ok)(acceptedProtocols) ? [...acceptedProtocols] : [];
    const p = (0, strings_1.$ftrim)(s).toLowerCase();
    if (p.length > 0 && p.match(__protocolCompleteRegex)) {
        const w3c = TSURL.StandardURLProtocols.has(p);
        if (w3c || _validateOtherProtocol(p, aps)) {
            return [p, w3c, aps];
        }
    }
    return ['', false, aps];
}
function _findProtocol(s) {
    const m = __protocolRegex.exec(s);
    return (0, commons_1.$ok)(m) ? m[0] : null;
}
function _parseHostAndPort(s) {
    s = (0, strings_1.$ascii)((0, strings_1.$ftrim)(s)).toLowerCase();
    const m = __portRegex.exec(s);
    if ((0, commons_1.$ok)(m)) {
        const port = m[0];
        return [s.slice(0, s.length - port.length), port.slice(1)];
    }
    return [s, ''];
}
function _validateOtherProtocol(protocol, acceptedProtocols) {
    let found = acceptedProtocols.find(a => {
        let p = (0, strings_1.$ftrim)(a).toLowerCase();
        if (!p.endsWith(':')) {
            p += ':';
        }
        return p === protocol;
    });
    return (0, commons_1.$ok)(found);
}
function _validateHostName(hostname) {
    const hostparts = hostname.split(/\./);
    const n = hostparts.length;
    for (let i = 0; i < n; i++) {
        var part = hostparts[i];
        if (!part.length) {
            continue;
        }
        if (!part.match(__hostnamePartRegex)) {
            var newPart = part.replace(__nonASCIIGlobalRegex, _ => 'x');
            if (!newPart.match(__hostnamePartRegex)) {
                const validParts = hostparts.slice(0, i);
                const invalidParts = hostparts.slice(i + 1);
                const m = part.match(__hostnamePartStartRegex);
                if ((0, commons_1.$ok)(m)) {
                    validParts.push(m[1]);
                    invalidParts.unshift(m[2]);
                }
                return [validParts.join('.'), invalidParts.length ? '/' + invalidParts.join('.') : ''];
            }
        }
    }
    return [hostname, ''];
}
//# sourceMappingURL=tsurl.js.map